/*
	MixCmd.cc

	Will Eighteen, 2000, 2006: will@opaline.demon.co.uk

	note: should mix to padded track length of (75?) byte multiples for CDA format?  
*/

#include "OpResources.h"
#include "MixCmd.h"
#include "PakDefs.h"
#include <MessageDialogManager.h>
#include <math.h>
#include <FileAccess.h>


MixCmd::MixCmd ( char *name, int active, char *mixName ) : BreakCmd ( name, active )
{
	_mixFileName = mixName;
	_mixMagnitude = 0;
	_writeSamples = 0;
	_processedSamples = 0;
	_mixBytes = 0;
	_mixBlockLength = mixwaveResources->blockLength();
	_pMixBlock = 0;
	_gainArray = 0;
	_panArray = 0;
	_halfPi = 1.5707963;
	_mixBits = 32; // default mixdown array size
	// the mixer requires all loaded waves to have equal bit resolution
	_bytesPerSample = mixwaveResources->bitsPerSample() / 8;
	_maxDataLength = 0;
	_once = 0;
	_pMixed = 0;	// no output wave
	_error[0] = 0;
	_primaryMixDone = False;
	_numWaves = 0;
	_currentWaveNum = 0;
	_lostBits = 0;
	_start = 0;	// start and end of earliest and latest selections respectively
	_end = 0;
	_sampleCount = 0;
	updateTimeout ( 100 );
}


MixCmd::~MixCmd()
{
	// empty
}


void MixCmd::doCommand()
{
	int percent;
	char text[18];

	if ( _primaryMixDone == False )
	{
		if ( !_once )
		{
			_numWaves = wavePak->Count();

			if ( preMix() )
			{
				_done = True;
				postMix();
				return;
			}

			_maxDataLength = ( _end - _start ) << ( _bytesPerSample - 1 );

			_mixBytes = _maxDataLength;
			_pMixed = new Wave;
			_pMixed->AllocateDataSpace ( _mixBlockLength << 1 );	// stereo mixdown
			_oData = _pMixed->data();

			if ( !_pMixed->SaveWav ( _mixFileName ) )
			{
				char msg[1024];
				snprintf ( msg, sizeof ( msg ), "%s\nSave failed: cannot create file.", _mixFileName );
				msg [ sizeof ( msg ) - 1 ] = 0;
				messageDialogManager->post ( msg );
				postMix();
				return;
			}

			_once = 1;
		}

		if ( _mixBytes < _mixBlockLength )
			_mixBlockLength = _mixBytes;

		_mixData = mix ( _mixBlockLength );

		if ( !_mixData )
		{
			char msg[80];
			snprintf ( msg, sizeof ( msg ), "Mixdown error: block mix failed after %ld of %ld samples.",
						_processedSamples, _mixBytes >> ( _bytesPerSample - 1 ) );
			msg [ sizeof ( msg ) - 1 ] = 0;
			messageDialogManager->post ( msg );
			postMix();
			_done = True;
			return;
		}
		else
		{
			// returned data is in 32 bits to preserve digits
			for ( unsigned int count = 0; count < _writeSamples << ( _bytesPerSample - 1 ); count ++ )
			{	// stuff data in 16-bit form in output array
				// alter this bit to account for mixBits
				// discard top 16 bits, can't use these... 
				// needs to account for 8-bit data as well

				// build up L and R
				_oData[count<<1] = _mixData[count]&0x000000ff;
				_oData[(count<<1)+1] = _mixData[count]>>8;
			}

			_pMixed->FillHeader ( 2, mixwaveResources->samplesPerSec(),
									mixwaveResources->bitsPerSample(), _processedSamples );

			_pMixed->AppendBlock ( _mixFileName, 2, mixwaveResources->samplesPerSec(),
									mixwaveResources->bitsPerSample(), _writeSamples );

			// overwrite header with modified data length                   
			_pMixed->ChangeHeader ( _mixFileName );
			_mixBytes -= _writeSamples << ( _bytesPerSample - 1 );
			freeMixBlock();
			_sampleCount += _writeSamples;
		}

		if ( _mixBytes == 0 )	// end primary mixdown, truncated to LS byte
		{
//			unsigned int waveFormat;

			_primaryMixDone = True;
/*
			if ( mixwaveResources->minY() < 0 )
				waveFormat = 1;	// signed
			else
				waveFormat = 0;
*/
			if ( mixwaveResources->minY() < 0 )	// signed data,max. half _mixBits magnitude
			{			
				if ( maxBitSet ( _mixMagnitude ) > (_pMixed->bitsPerSample() - 1 ) )
					_lostBits = maxBitSet ( _mixMagnitude ) - _pMixed->bitsPerSample() + 1;
			}
			else
			{
				if ( maxBitSet ( _mixMagnitude ) > _pMixed->bitsPerSample() )
					_lostBits = maxBitSet ( _mixMagnitude ) - _pMixed->bitsPerSample();
			}

			if ( mixwaveResources->enableRemix() == True )
			{
				_mixBlockLength = mixwaveResources->blockLength();

				if ( _lostBits > _mixBits)
				{
					strncpy ( _error, "Remix error: overrange: reduce gain. ", sizeof ( _error ) );
					_error [ sizeof ( _error ) - 1 ] = 0;
				}
				else
				{
					if ( _lostBits > 0 )
					{
						char intText[20];
						strncpy ( _error, "Mixdown: right-shifted mix ", sizeof ( _error ) );		
						_error [ sizeof ( _error ) - 1 ] = 0;
						snprintf ( intText, sizeof ( intText ), "%lu", _lostBits );
						intText [ sizeof ( intText ) - 1 ] = 0;
						strncat ( _error, intText, sizeof ( _error ) - strlen ( _error ) - 1 );
						_error [ sizeof ( _error ) - 1 ] = 0;

						if ( _lostBits > 1 )
    	   					strncat ( _error, " bits.", sizeof ( _error ) - strlen ( _error ) - 1 );
						else
							strncat ( _error, " bit.", sizeof ( _error ) - strlen ( _error ) - 1 );

						_error [ sizeof ( _error ) - 1 ] = 0;

						_processedSamples = 0;
						_percentDone = 0;
						_mixBytes = _maxDataLength;

						reInitWavePointers();

						_pMixed->filePos = 0;
						_pMixed->FillHeader ( 2, mixwaveResources->samplesPerSec(),
												mixwaveResources->bitsPerSample(), _processedSamples );
						_pMixed->SaveHeader ( _mixFileName );	// overwrite existing header so append begins at start
					}
				}
			}
			else	// enableRemix is false
			{
				if ( _lostBits > 0 )
				{
					strncpy ( _error, "Mixdown error: overrange: reduce gain. ", sizeof ( _error ) );
					_error [ sizeof ( _error ) - 1 ] = 0;
				}
			}
		}
	}	// end primary mixdown, truncated to LSBits (lower 16 of 32 )
	
	// remix if primary mix bits truncated 
	if ( ( _primaryMixDone == True ) && ( mixwaveResources->enableRemix() == True ) && ( _lostBits > 0 ) )
	{	
		if ( _mixBytes < _mixBlockLength )
			_mixBlockLength = _mixBytes;

		_mixData = mix ( _mixBlockLength );

		if ( !_mixData )
		{
			char msg[80];
			snprintf ( msg, sizeof ( msg ), "Mixdown error: block remix failed after %ld of %ld samples.",
						_processedSamples, _mixBytes >> ( _bytesPerSample - 1 ) );
			msg [ sizeof ( msg ) - 1 ] = 0;
			messageDialogManager->post ( msg );
			postMix();
			_done = True;
			return;
		}
		else
		{
			// returned data is in 32 bits to preserve digits
			for ( unsigned int count = 0; count < _writeSamples << ( _bytesPerSample - 1 ); count ++ )
			{
				_mixData[count] = _mixData[count]>>_lostBits;
				_oData[count<<1] = _mixData[count]&0x000000ff;
				_oData[(count<<1)+1] = _mixData[count]>>8;
			}	// avoiding 'thin' mixes

			_pMixed->FillHeader ( 2, mixwaveResources->samplesPerSec(),
									mixwaveResources->bitsPerSample(), _processedSamples );

			_pMixed->AppendBlock ( _mixFileName, 2, mixwaveResources->samplesPerSec(),
									mixwaveResources->bitsPerSample(), _writeSamples );

			// overwrite header with modified data length                   
			_pMixed->ChangeHeader ( _mixFileName );
			_mixBytes -= _writeSamples << ( _bytesPerSample - 1 );			
			freeMixBlock();
			_sampleCount += _writeSamples;
		}
	}	// end remix, LS bits truncated to make MSbits 16-bit output

	percent = ( int ) ( ( float ) _sampleCount / ( float ) ( _maxDataLength	>> ( _bytesPerSample - 1 ) ) * 100 );

	if ( _percentDone != percent )
	{
		if ( percent > 100 )
			percent = percent - 100;

		_percentDone = percent;

		if ( ( mixwaveResources->enableRemix() == True ) && ( _primaryMixDone == True ) )
			snprintf ( text, 17, "Remixing... %d%%", _percentDone );
		else
			snprintf ( text, 5, "%d%%", _percentDone );

		updateMessage ( text );
	}

	if ( ( _mixBytes == 0 ) && ( _primaryMixDone == True ) )	// no data left to mix
	{
		postMix();
		_done = True;
	}
}


int MixCmd::preMix()
{
	int n, error;
	char msg[BUFSIZE];
	
	error = 0;
	
	for ( n = 0; n < _numWaves; n++ )
	{
		if ( (*wavePak)[n].fLock == ACTIVE )
		{
			if ( mixwaveResources->enableWarningPopups() == True )
			{
				snprintf ( msg, sizeof ( msg ), "%s\nFile in use!", (*wavePak)[n].pWave->name() );
				msg [ sizeof ( msg ) - 1 ] = 0;
				messageDialogManager->post ( msg );
			}

			error = 1;
		}
		else
		{
			(*wavePak)[n].fLock = ACTIVE;
			(*wavePak)[n].pWave->filePos = 0;

			if ( (*wavePak)[n].gainTemplate != 0 )	// i.e. there is one
				(*wavePak)[n].gainRatio = ( float ) ( (*wavePak)[n].pWave->totalSamples() - 1 ) / ( float ) ( (*templatePak)[(*wavePak)[n].gainTemplate-1].dataSize - 1 );

			if ( (*wavePak)[n].panTemplate != 0 )
				(*wavePak)[n].panRatio = ( float ) ( (*wavePak)[n].pWave->totalSamples() - 1 ) / ( float ) ( (*templatePak)[(*wavePak)[n].panTemplate-1].dataSize - 1 );

			(*wavePak)[n].fLock = ACTIVE;
		}
	}
	
	for ( n = 0; n < templatePak->Count(); n++ )
	{
		if ( (*templatePak)[n].fLock == ACTIVE )
		{
			if ( mixwaveResources->enableWarningPopups() == True )
			{
				snprintf ( msg, sizeof ( msg ), "%s\nFile in use!", (*templatePak)[n].name );
				msg [ sizeof ( msg ) - 1 ] = 0;
				messageDialogManager->post ( msg );
			}

			error = 1;
		}
		else
			(*templatePak)[n].fLock = ACTIVE;
	}

	if ( error ) 
	{
		strncpy ( _error, "Mixdown error: premix terminated.", sizeof ( _error ) );
		_error [ sizeof ( _error ) - 1 ] = 0;
		return ( error );
	}

	// waves files handled by wave class
	// open template files for read

	for ( n = 0; n < _numWaves; n++ )
	{
		if ( (*wavePak)[n].gainTemplate != 0 )
		{
			FileAccess *gain = new FileAccess ( (*templatePak)[(*wavePak)[n].gainTemplate-1].name, "r" );
			(*templatePak)[(*wavePak)[n].gainTemplate-1].file = gain;
		}
		
		if ( (*wavePak)[n].panTemplate != 0 )
		{
			FileAccess *pan = new FileAccess ( (*templatePak)[(*wavePak)[n].panTemplate-1].name, "r" );
			(*templatePak)[(*wavePak)[n].panTemplate-1].file = pan;
		}
	}

	// find max data size
	_end = 0;

	for ( n = 0; n < _numWaves; n++ )
	{
		if ( (*wavePak)[n].opMode == 1 )
		{
			if ( (*wavePak)[n].endSel > _end )
				_end = (*wavePak)[n].endSel;
		}
		else
		{
			if ( (*wavePak)[n].pWave->totalSamples() > _end )
				_end = (*wavePak)[n].pWave->totalSamples();
		}
	}

	_start = _end;	// this is _total_ data between earliest/lastest selns

	for ( n = 0; n < _numWaves; n++ )
	{
		if ( (*wavePak)[n].opMode == 1 )
		{
			if ( (*wavePak)[n].startSel < _start )
				_start = (*wavePak)[n].startSel;
		}
		else
			_start = 0;
	}

	// find current wave num
	for ( n = 0; n < _numWaves; n++ )
		if ( strcmp ( (*wavePak)[n].pWave->name(), mixwaveResources->currentWave() ) == 0 )
			_currentWaveNum = (*wavePak)[n].pWave->waveNum();

	if ( initMixPakPointers() == 1 )
		return ( 1 );
	else
		return ( 0 );

}


void MixCmd::postMix()
{
	int count;
	
	for ( count = 0; count < _numWaves; count++ )
	{
		if ( (*wavePak)[count].gainTemplate != 0 )
		{
			FileAccess *gain = (*templatePak)[(*wavePak)[count].gainTemplate-1].file;
			delete gain;
		}

		if ( (*wavePak)[count].panTemplate != 0 )
		{
			FileAccess *pan = (*templatePak)[(*wavePak)[count].panTemplate-1].file;
			delete pan;
		}
		
		(*wavePak)[count].fLock = INACTIVE;
	}

	for ( count = 0; count < templatePak->Count(); count++ )
		(*templatePak)[count].fLock = INACTIVE;

	if ( _pMixBlock )
		freeMixBlock();	// this shouldn't get called...

	delete[] _gainArray;
	delete[] _panArray;

	if ( _pMixed )
		delete _pMixed;
}


unsigned long MixCmd::maxBitSet ( unsigned long mixMagnitude )
{ 
	unsigned long maxBitSet = 0;
	unsigned int count, testBits;
	
	testBits = _mixBits;

	if ( mixwaveResources->minY() < 0 )
		testBits--;

	for ( count = 0; count < ( testBits  ); count++ )
	{
		maxBitSet = ( unsigned long ) pow ( 2, count );

		if ( maxBitSet >= mixMagnitude )
		{
			maxBitSet = count;
			count = testBits - 1;
		}
	}

	return maxBitSet;
}


long *MixCmd::mix ( long blockLength )
{
	// NOTE: the Mix returns mixBits-bits data
	char *cData;
	long count;
	unsigned int i, j, k;
	long track_val;
	double mix_l, mix_r;
	Wave *pCurrentWave;
	Boolean skipWave;

	if ( !allocMixBlock ( blockLength ) )
	{
		// failed to allocate the long integer data block
		return ( 0 );
	}

	for ( count = 0; count < blockLength; count++ )
		_pMixBlock[count] = 0;

	for ( pCurrentWave = pCurrentWave->first(); pCurrentWave; pCurrentWave = pCurrentWave->next() )
	{
		skipWave = False;

		// if current wave selection opMode True, use selection, else whole wave
		if ( pCurrentWave->name() != NULL )	// skip just-created o/p file
		{
			if ( ( ( pCurrentWave->filePos + blockLength ) >> ( _bytesPerSample - 1 ) ) < (*mixPak)[pCurrentWave->waveNum()-1].start )
				skipWave = True;
			else
				if ( ( pCurrentWave->filePos >> ( _bytesPerSample - 1 ) ) >= (*mixPak)[pCurrentWave->waveNum()-1].end )
					skipWave = True;

			if ( skipWave == False )
			{
				if ( pCurrentWave->LoadBlock ( pCurrentWave->filePos, blockLength ) )
				{	// don't bother mixing if no samples!   
					loadGainArray ( pCurrentWave->waveNum(), blockLength>>( _bytesPerSample - 1 ) );
					loadPanArray ( pCurrentWave->waveNum(), blockLength>>( _bytesPerSample - 1 ) );
					cData = pCurrentWave->data();

					if ( ( mixwaveResources->mixToCurrent() == False ) && ( (*wavePak)[pCurrentWave->waveNum()-1].opMode == 1 ) )
					{
						if ( (*mixPak)[pCurrentWave->waveNum()-1].start > ( pCurrentWave->filePos >> ( _bytesPerSample - 1 ) ) )
						{	// seln start > block start
							int diff;

							diff = ( (*mixPak)[pCurrentWave->waveNum()-1].start
									- ( pCurrentWave->filePos >> ( _bytesPerSample - 1 ) ) )
									<< ( _bytesPerSample - 1 );

							for ( count = 0; count < diff; count++ )
								cData[count] = 0;
						}

						if ((*mixPak)[pCurrentWave->waveNum()-1].end < ( (pCurrentWave->filePos + blockLength ) >> ( _bytesPerSample - 1 )) )
						{	// seln end < block end
							int diff;

							diff = ( (*mixPak)[pCurrentWave->waveNum()-1].end
									<< ( _bytesPerSample - 1 ) ) - pCurrentWave->filePos;

							for ( count = diff; count < blockLength; count++ )
								cData[count] = 0;
						}
					}

					for ( count = 0; count < blockLength; count += _bytesPerSample )
					{
						switch ( _bytesPerSample )
						{
							case 1:
								break;
								
							case 2:
								i = ( unsigned char ) cData[count];
								j = ( unsigned char ) cData[count+1];
								track_val = i + 256 * j;

								// limit val to maxY()+1 if signed, otherwise maxY()
								if ( mixwaveResources->minY() < 0 )
								{
									if ( track_val > mixwaveResources->maxY() )
										track_val = ( track_val - ( 2 * mixwaveResources->maxY() + 1 ) ) - 1;
								}
								else
									if ( track_val > mixwaveResources->maxY() )
										track_val = mixwaveResources->maxY();

								mix_l = track_val * cos ( _panArray[count>>1] * _halfPi ) * _gainArray[count>>1];
								mix_r = track_val * sin ( _panArray[count>>1] * _halfPi ) * _gainArray[count>>1];
								break;
							
							case 3:
								i = ( unsigned char ) cData[count];
								j = ( unsigned char ) cData[count+1];
								k = ( unsigned char ) cData[count+2];
								track_val = i + 256 * j + 256^2 * k;
								break;
						}

						_pMixBlock[count] += ( long ) mix_l;		// build up L and R
						_pMixBlock[count + 1] += ( long ) mix_r;	// composite values

						// find max. value in mix, hence max. digits used
						if ( abs ( _pMixBlock[count] ) > abs ( _mixMagnitude ) )
							_mixMagnitude = abs ( _pMixBlock[count] );

						if ( abs ( _pMixBlock[count + 1] ) > abs ( _mixMagnitude ) )
							_mixMagnitude = abs ( _pMixBlock[count + 1] );
					}
				}
			}  // end for current wave

			pCurrentWave->filePos += blockLength;
		}	// end check current wave valid
	} // end for all waves

	_processedSamples += blockLength>>( _bytesPerSample - 1 );      
	_writeSamples = blockLength>>( _bytesPerSample - 1 );

	return _pMixBlock;	// return pointer to block mixed from input blocks
}


Boolean MixCmd::allocMixBlock ( long size )
{
	freeMixBlock();

	_pMixBlock = new long [size];

	if ( _pMixBlock )
		return True;
	else
		return False;
}


void MixCmd::freeMixBlock()
{
	if ( _pMixBlock )
		delete[] _pMixBlock;

	_pMixBlock = 0;
}

// note: load arrays methods take no note of case of large number of template
// data corres. to small wave data. Then arrays may be huge; should load
// these by resource-blocklength-sized chunks, perhaps.

void MixCmd::loadGainArray ( unsigned int waveNum, unsigned long numData )
{
	unsigned long numSamples;
	float *gainArrayT;
	unsigned int gainTemplatePak;
	unsigned long count, dataPoints, dstArrayPtr, i, j;
	int mode;
	float increment, lValue, ratio, rValue;

	mode = 0;	// default for ratio = 1
	ratio = 1;
	dataPoints = 1;
	
	if ( (*wavePak)[waveNum-1].gainTemplate != 0 )	// i.e. there is one
	{
		int waveSample = (*wavePak)[waveNum-1].pWave->filePos >> ( _bytesPerSample - 1 );
		gainTemplatePak = (*wavePak)[waveNum-1].gainTemplate - 1;	
		ratio = (*wavePak)[waveNum-1].gainRatio;
		int tSample = ( int ) ( ( float ) waveSample / ratio );

		if ( ratio < 1.0 )	// numWaveDataPoints < numTemplateDataPoints
		{
			mode = -1;
			dataPoints = ( int ) ( 1 / ratio );
		}
		else
		{
			if ( ratio > 1.0 )	// numWaveDataPoints > numTemplateDataPoints
			{
				mode = 1;
				dataPoints = ( int ) ratio;	// num points between template samples
			}
		}
		
		(*templatePak)[gainTemplatePak].file->filePos ( tSample * sizeof ( float ) );
	}

	numSamples = ( long ) ( ( float ) numData / ratio );

	if  ( _gainArray )
	{
		delete[] _gainArray;
		_gainArray = 0;
	}
	
	if ( numSamples <= 1 )
		gainArrayT = new float [ 1 ];
	else
		gainArrayT = new float [ numSamples ];

	_gainArray = new float [ numData ];

	if ( (*wavePak)[waveNum-1].gainTemplate != 0 )
	{
		// load from template gainTemplatePak
		(*templatePak)[gainTemplatePak].file->elementSize ( sizeof ( float ) );
		
		if ( numSamples == 0 )
			(*templatePak)[gainTemplatePak].file->numElements ( 1 );	
		else
			(*templatePak)[gainTemplatePak].file->numElements ( numSamples );	

		(*templatePak)[gainTemplatePak].file->readF ( gainArrayT );

		switch ( mode )
		{
			case -1:	// average template data
				i = 0;
				dstArrayPtr = 0;
				float mean;

				while ( i < ( numSamples - dataPoints ) )
				{
					mean = 0;

					for ( j = i; j < ( i + dataPoints ); j++ )
						mean += gainArrayT[j];
						
					mean /= dataPoints;
					_gainArray[dstArrayPtr] = mean;

					if ( ( ( unsigned long ) ( ( float ) i * ratio ) ) > dstArrayPtr )
						dataPoints--;
					else
						if ( ( ( unsigned long ) ( ( float ) i * ratio ) ) < dstArrayPtr )
							dataPoints++;

					dstArrayPtr++;						
					i += dataPoints;
				}

				if ( i < numSamples )
				{
					j = 0;
					i -= ( ( int ) ( 1 / ratio ) );
					mean = 0;
					count = numSamples - i;

					while ( i < numSamples )
					{
						mean += gainArrayT[i];
						i++;
					}

					mean /= count;
					_gainArray[numData-1] = mean;
				}

				break;

			case 0:	// 1-to-1 correspondence
				for ( i = 0; i < numData; i++ )
					_gainArray[i] = gainArrayT[i];

				break;

			case 1:	// interpolate template data
				count = 0;
				i = 0;
				dstArrayPtr = 0;
				
				if ( numSamples <= 1 )
				{				
					while ( i < numData )
					{
						_gainArray[i] = gainArrayT[count];
						i++;
					}
				}
				else
				{
					while ( count < ( numSamples - 1 ) )
					{
						lValue = gainArrayT[count];
						rValue = gainArrayT[count+1];
						increment = rValue - lValue;
					
						if ( increment != 0 )
							increment /= ratio;

						i = 0;
						j = 0;

						while ( i < dataPoints )
						{
							_gainArray[dstArrayPtr+i] = lValue + ( j * increment );
							i++;
							j++;
						}

						count ++;
						dstArrayPtr += dataPoints;
					}

					count = dstArrayPtr;
					lValue = _gainArray[dstArrayPtr-1];

					if ( ( (*templatePak)[gainTemplatePak].file->filePos()>>sizeof ( float ) ) != numData )
					{
						(*templatePak)[gainTemplatePak].file->numElements ( 1 );
						(*templatePak)[gainTemplatePak].file->readF ( &rValue );
						(*templatePak)[gainTemplatePak].file->filePos( (*templatePak)[gainTemplatePak].file->filePos() - sizeof ( float ) );
					}	// interpolate between end of array and start of next array

					i = 1;
					increment = ( rValue - lValue ) / ( ( ( float ) numData - ( float ) count ) + 1 );

					while ( count < numData )	// pad to end of gain array
					{
						_gainArray[count] = lValue + ( i * increment );
						count++;
						i++;
					}
				}
				
				break;
		}		
	}
	else
	{
		// calculate fixed data from pak gain
		for ( i = 0; i < numData; i++ )
			_gainArray[i] = (*wavePak)[waveNum-1].gain;
	}

	delete[] gainArrayT;
}


void MixCmd::loadPanArray ( unsigned int waveNum, unsigned long numData )
{
	unsigned long numSamples, count, dataPoints, dstArrayPtr, i, j;
	float *panArrayT, increment, lValue, ratio, rValue;
	unsigned int panTemplatePak;
	int mode;

	mode = 0;	// default for ratio = 1
	ratio = 1;
	dataPoints = 1;
	
	if ( (*wavePak)[waveNum-1].panTemplate != 0 )	// i.e. there is one
	{
		int waveSample = (*wavePak)[waveNum-1].pWave->filePos >> ( _bytesPerSample - 1 );
		panTemplatePak = (*wavePak)[waveNum-1].panTemplate - 1;	
		ratio = (*wavePak)[waveNum-1].panRatio;
		int tSample = ( int ) ( ( float ) waveSample / ratio );

		if ( ratio < 1.0 )	// numWaveDataPoints < numTemplateDataPoints
		{
			mode = -1;
			dataPoints = ( int ) ( 1 / ratio );
		}
		else
		{
			if ( ratio > 1.0 )	// numWaveDataPoints > numTemplateDataPoints
			{
				mode = 1;
				dataPoints = ( int ) ratio;	// num points between template samples
			}
		}
		
		(*templatePak)[panTemplatePak].file->filePos ( tSample * sizeof ( float ) );
	}

	numSamples = ( long ) ( ( float ) numData / ratio );

	if  ( _panArray )
	{
		delete[] _panArray;
		_panArray = 0;
	}
	
	if ( numSamples <= 1 )
		panArrayT = new float [ 1 ];
	else
		panArrayT = new float [ numSamples ];

	_panArray = new float [ numData ];

	if ( (*wavePak)[waveNum-1].panTemplate != 0 )
	{
		// load from template panTemplatePak
		(*templatePak)[panTemplatePak].file->elementSize ( sizeof ( float ) );
		
		if ( numSamples == 0 )
			(*templatePak)[panTemplatePak].file->numElements ( 1 );	
		else
			(*templatePak)[panTemplatePak].file->numElements ( numSamples );	

		(*templatePak)[panTemplatePak].file->readF ( panArrayT );

		switch ( mode )
		{
			case -1:	// average template data
				i = 0;
				dstArrayPtr = 0;
				float mean;

				while ( i < ( numSamples - dataPoints ) )
				{
					mean = 0;

					for ( j = i; j < ( i + dataPoints ); j++ )
						mean += panArrayT[j];
						
					mean /= dataPoints;
					_panArray[dstArrayPtr] = mean;

					if ( ( ( unsigned long ) ( ( float ) i * ratio ) ) > dstArrayPtr )
						dataPoints--;
					else
						if ( ( ( unsigned long ) ( ( float ) i * ratio ) ) < dstArrayPtr )
							dataPoints++;

					dstArrayPtr++;						
					i += dataPoints;
				}

				if ( i < numSamples )
				{
					j = 0;
					i -= ( ( unsigned long ) ( 1 / ratio ) );
					mean = 0;
					count = numSamples - i;

					while ( i < numSamples )
					{
						mean += panArrayT[i];
						i++;
					}

					mean /= count;
					_panArray[numData-1] = mean;
				}

				break;

			case 0:	// 1-to-1 correspondence
				for ( i = 0; i < numData; i++ )
					_panArray[i] = panArrayT[i];

				break;

			case 1:	// interpolate template data
				count = 0;
				i = 0;
				dstArrayPtr = 0;
				
				if ( numSamples <= 1 )
				{				
					while ( i < numData )
					{
						_panArray[i] = panArrayT[count];
						i++;
					}
				}
				else
				{
					while ( count < ( numSamples - 1 ) )
					{
						lValue = panArrayT[count];
						rValue = panArrayT[count+1];
						increment = rValue - lValue;
					
						if ( increment != 0 )
							increment /= ratio;

						i = 0;
						j = 0;

						while ( i < dataPoints )
						{
							_panArray[dstArrayPtr+i] = lValue + ( j * increment );
							i++;
							j++;
						}

						count ++;
						dstArrayPtr += dataPoints;
					}

					count = dstArrayPtr;
					lValue = _panArray[dstArrayPtr-1];

					if ( ( (*templatePak)[panTemplatePak].file->filePos()>>sizeof ( float ) ) != numData )
					{
						(*templatePak)[panTemplatePak].file->numElements ( 1 );
						(*templatePak)[panTemplatePak].file->readF ( &rValue );
						(*templatePak)[panTemplatePak].file->filePos( (*templatePak)[panTemplatePak].file->filePos() - sizeof ( float ) );
					}	// interpolate between end of array and start of next array

					i = 1;
					increment = ( rValue - lValue ) / ( ( ( float ) numData - ( float ) count ) + 1 );

					while ( count < numData )	// pad to end of gain array
					{
						_panArray[count] = lValue + ( i * increment );
						count++;
						i++;
					}
				}
				
				break;
		}		
	}
	else
	{
		// calculate fixed data from pak pan
		for ( i = 0; i < numData; i++ )
			_panArray[i] = (*wavePak)[waveNum-1].pan;
	}

	delete[] panArrayT;
}


int MixCmd::initMixPakPointers ( void )
{
	// set up mixPak pointers for waves
	// if mixToCurrent true, check wave specified, then check for selection
	int n, m;

	if ( mixwaveResources->syncSelections() == True )
	{
		for ( n = 0; n < _numWaves; n++ )
		{
			if ( (*wavePak)[n].startSel == (*wavePak)[n].endSel )
			{
				{
					snprintf ( _error, sizeof ( _error ), "Mixdown error: request to mix to selection \
in wave\n%s\nError: no selection in wave.", (*wavePak)[n].pWave->name() );
					_error [ sizeof ( _error ) - 1 ] = 0;
					return ( 1 );
				}
			}
		}

		_end = 0;

		for ( n = 0; n < _numWaves; n++ )
		{
			if ( (*wavePak)[n].endSel - (*wavePak)[n].startSel > _end )
				_end = (*wavePak)[n].endSel - (*wavePak)[n].startSel;
		}

		for ( n = 0; n < _numWaves; n++ )
		{
			(*mixPak)[n].start = (*wavePak)[n].startSel;
			(*mixPak)[n].end = (*wavePak)[n].endSel;
			(*wavePak)[n].pWave->filePos = (*mixPak)[n].start << ( _bytesPerSample - 1 );
		}

		return ( 0 );
	}

	if ( mixwaveResources->mixToCurrent() == True )
	{
		if ( _currentWaveNum == 0 )
		{
			snprintf ( _error, sizeof ( _error ), "Mixdown error: premix terminated.\n\
No current wave set in Options / Wave control" );
			_error [ sizeof ( _error ) - 1 ] = 0;
			return ( 1 );
		}

		if ( (*wavePak)[_currentWaveNum-1].opMode == 0 )
		{
			for ( n = 0; n < _numWaves; n++ )
			{	// mix all
				(*mixPak)[n].start = 0;
				(*mixPak)[n].end = (*wavePak)[n].pWave->totalSamples();
				(*wavePak)[n].pWave->filePos = 0;
			}
		}
		else
		{
			if ( (*wavePak)[_currentWaveNum-1].startSel == (*wavePak)[_currentWaveNum-1].endSel )
			{
				snprintf ( _error, sizeof ( _error ), "Mixdown error: request to mix to selection \
in wave\n%s\nError: no selection in wave.", (*wavePak)[_currentWaveNum-1].pWave->name() );
				_error [ sizeof ( _error ) - 1 ] = 0;
				return ( 1 );
			}
			else
			{	// make selections for mix based on currentWave selection
				for ( n = 0; n < _numWaves; n++ )
				{
					(*mixPak)[n].start = (*wavePak)[_currentWaveNum-1].startSel;
					(*mixPak)[n].end = (*wavePak)[_currentWaveNum-1].endSel;
					// set file pointers to common start
					(*wavePak)[n].pWave->filePos = (*mixPak)[n].start << ( _bytesPerSample - 1 );
				}

				_start = (*mixPak)[_currentWaveNum-1].start;
				_end = (*mixPak)[_currentWaveNum-1].end;
			}
		}
	}
	else
	{	// mixToCurrent False
		for ( n = 0; n < _numWaves; n++ )
		{
			for ( m = 0; m < _numWaves; m++ )
			{
				if ( (*wavePak)[n].pWave->waveNum() == (*mixPak)[m].waveNum )
				{
					if ( (*wavePak)[n].opMode == 1)
					{	// mix selection from wave
						(*mixPak)[m].start = (*wavePak)[n].startSel;
						(*mixPak)[m].end = (*wavePak)[n].endSel;
					}

					else
					{	// default to 'mix all' if no selections
						(*mixPak)[m].start = 0;
						(*mixPak)[m].end = (*wavePak)[n].pWave->totalSamples();
					}

					// set all fileptrs to start of earliest selection or 0 if whole wave
					(*wavePak)[n].pWave->filePos = _start << ( _bytesPerSample - 1 );
	
					m = _numWaves;
				}
			}
		}
	}
	
	return ( 0 );
}


void MixCmd::reInitWavePointers ( void )
{
	unsigned int n;

	if ( mixwaveResources->syncSelections() == True )
	{
		for ( n = 0; n < _numWaves; n++ )
			(*wavePak)[n].pWave->filePos = (*mixPak)[n].start << ( _bytesPerSample - 1 );

		return;
	}

	if ( mixwaveResources->mixToCurrent() == True )
	{
		if ( (*wavePak)[_currentWaveNum-1].opMode == 0 )
		{
			for ( n = 0; n < _numWaves; n++ )
				(*wavePak)[n].pWave->filePos = 0;
		}
		else
		{
			for ( n = 0; n < _numWaves; n++ )
				(*wavePak)[n].pWave->filePos = (*mixPak)[n].start << ( _bytesPerSample - 1 );
		}
	}
	else
		for ( n = 0; n < _numWaves; n++ )
			(*wavePak)[n].pWave->filePos = _start << ( _bytesPerSample - 1 );

	return;
}
